@@ -1,6 +1,7 @@
 import asyncio
 import asyncio.events as events
 import sys
+import threading
 from heapq import heappop
 
 
@@ -16,6 +17,7 @@ def apply(loop=None):
     _patch_loop(loop)
     _patch_task()
     _patch_handle()
+    _patch_tornado()
 
 
 def _patch_asyncio():
@@ -44,19 +46,55 @@ def _patch_asyncio():
 def _patch_loop(loop):
     """Patch loop to make it reentrent."""
 
-    def run_until_complete(self, future):
+    def run_forever(self):
+        if sys.version_info >= (3, 7, 0):
+            set_coro_tracking = self._set_coroutine_origin_tracking
+        else:
+            set_coro_tracking = self._set_coroutine_wrapper
+
         self._check_closed()
-        events._set_running_loop(self)
-        f = asyncio.ensure_future(future, loop=self)
-        if f is not future:
-            f._log_destroy_pending = False
-        while not f.done():
-            self._run_once()
-            if self._stopping:
-                break
-        if not f.done():
-            raise RuntimeError('Event loop stopped before Future completed.')
-        return f.result()
+        old_thread_id = self._thread_id
+        old_running_loop = events._get_running_loop()
+        set_coro_tracking(self._debug)
+        self._thread_id = threading.get_ident()
+
+        if self._asyncgens is not None:
+            old_agen_hooks = sys.get_asyncgen_hooks()
+            sys.set_asyncgen_hooks(
+                firstiter=self._asyncgen_firstiter_hook,
+                finalizer=self._asyncgen_finalizer_hook)
+        try:
+            events._set_running_loop(self)
+            while True:
+                self._run_once()
+                if self._stopping:
+                    break
+        finally:
+            self._stopping = False
+            self._thread_id = old_thread_id
+            events._set_running_loop(old_running_loop)
+            set_coro_tracking(False)
+            if self._asyncgens is not None:
+                sys.set_asyncgen_hooks(*old_agen_hooks)
+
+    def run_until_complete(self, future):
+        old_running_loop = events._get_running_loop()
+        try:
+            self._check_closed()
+            events._set_running_loop(self)
+            f = asyncio.ensure_future(future, loop=self)
+            if f is not future:
+                f._log_destroy_pending = False
+            while not f.done():
+                self._run_once()
+                if self._stopping:
+                    break
+            if not f.done():
+                raise RuntimeError(
+                    'Event loop stopped before Future completed.')
+            return f.result()
+        finally:
+            events._set_running_loop(old_running_loop)
 
     def _run_once(self):
         """
@@ -92,10 +130,12 @@ def _patch_loop(loop):
     cls = loop.__class__
     cls._run_once_orig = cls._run_once
     cls._run_once = _run_once
+    cls._run_forever_orig = cls.run_forever
+    cls.run_forever = run_forever
     cls._run_until_complete_orig = cls.run_until_complete
     cls.run_until_complete = run_until_complete
-    cls._check_running_orig = cls._check_running
     cls._check_running = _check_running
+    cls._check_runnung = _check_running  # typo in Python 3.7 source
     cls._nest_patched = True
 
 
@@ -119,7 +159,7 @@ def _patch_task():
             curr_tasks[loop] = task
 
         def leave_task(loop, task):
-            del curr_tasks[loop]
+            curr_tasks.pop(loop, None)
 
         asyncio.tasks._enter_task = enter_task
         asyncio.tasks._leave_task = leave_task
@@ -150,7 +190,9 @@ def _patch_handle():
             ctx.run(self._callback, *self._args)
             if ctx:
                 self._context.run(update_from_context, ctx)
-        except Exception as exc:
+        except (SystemExit, KeyboardInterrupt):
+            raise
+        except BaseException as exc:
             cb = format_helpers._format_callback_source(
                 self._callback, self._args)
             msg = 'Exception in callback {}'.format(cb)
@@ -167,3 +209,15 @@ def _patch_handle():
     if sys.version_info >= (3, 7, 0):
         from asyncio import format_helpers
         events.Handle._run = run
+
+
+def _patch_tornado():
+        """
+        If tornado is imported before nest_asyncio, make tornado aware of
+        the pure-Python asyncio Future.
+        """
+        if 'tornado' in sys.modules:
+            import tornado.concurrent as tc
+            tc.Future = asyncio.Future
+            if asyncio.Future not in tc.FUTURES:
+                tc.FUTURES += (asyncio.Future,)
